# encoding: utf-8
from datetime import datetime as _datetime
from pdef import Type

try:
    # Python 2.7
    _long = long
    _string = unicode
except NameError:
    # Python 3
    _long = int
    _string = str


class Descriptor(object):
    '''Base type descriptor.'''
    def __init__(self, type0, pyclass):
        self.type = type0
        self._pyclass_supplier = _supplier(pyclass)
        self._pyclass = None

        self.is_primitive = self.type in Type.PRIMITIVE_TYPES
        self.is_data_type = self.type in Type.DATA_TYPES
        self.is_message = self.type == Type.MESSAGE
        self.is_mutable = self.type in Type.MUTABLE_TYPES

    def __str__(self):
        return str(self.type)

    @property
    def pyclass(self):
        pyclass = self._pyclass
        if pyclass:
            return pyclass

        self._pyclass = self._pyclass_supplier()
        return self._pyclass


class DataTypeDescriptor(Descriptor):
    default = None

    def __init__(self, type0, pyclass):
        super(DataTypeDescriptor, self).__init__(type0, pyclass)


class MessageDescriptor(DataTypeDescriptor):
    '''Message descriptor.'''
    def __init__(self, pyclass, base=None, discriminator_value=None, fields=None, subtypes=None):
        super(MessageDescriptor, self).__init__(Type.MESSAGE, pyclass)
        self.base = base

        self.declared_fields = tuple(fields) if fields else ()
        self.inherited_fields = base.fields if base else ()
        self.fields = self.inherited_fields + self.declared_fields
        self._field_tuples = None

        self.discriminator_value = discriminator_value
        self.discriminator = self._find_discriminator(self.fields)

        self._subtype_suppliers = tuple(_supplier(s) for s in subtypes) if subtypes else ()
        self._subtypes = None
        self._subtype_map = None

        self.is_polymorphic = bool(self.discriminator)

    def __str__(self):
        return str(self.pyclass)

    @property
    def default(self):
        return self.pyclass()

    @classmethod
    def _find_discriminator(cls, fields):
        for field in fields:
            if field.is_discriminator:
                return field

    @property
    def subtypes(self):
        subtypes = self._subtypes
        if subtypes:
            return subtypes

        self._subtypes = tuple(supplier() for supplier in self._subtype_suppliers)
        self._subtype_map = {s.discriminator_value: s for s in self._subtypes}
        return self._subtypes

    @property
    def field_tuples(self):
        field_tuples = self._field_tuples
        if field_tuples:
            return field_tuples

        self._field_tuples = tuple((field.name, field.private_name, field.type)
                                   for field in self.fields)
        return self._field_tuples

    def find_field(self, name):
        '''Return a field by its name or None.'''
        for field in self.fields:
            if field.name == name:
                return field

    def find_subtype(self, type0):
        '''Return a subtype descriptor by a type enum value or self if not found.'''
        _ = self.subtypes  # Force loading subtypes.

        subtype = self._subtype_map.get(type0)
        return subtype if subtype else self


class FieldDescriptor(object):
    '''Message field descriptor which provides a field meta-data and implements the python
    descriptor protocol (it is a property).

    Example as a property::
    >>> class A(object):
    >>>     hello = field('hello', type=lambda: string0)
    >>>     has_hello = hello.has_property
    >>> a = A()
    >>> a.field = 'hello'

    Property equivalent::
    >>> class A(object):
    >>>     @property
    >>>     def hello(self):
    >>>         return self._hello or ''  # or the default value.
    >>>     @hello.setter
    >>>     def hello(self, value):
    >>>         self._hello = value
    >>>     @property
    >>>     def has_hello(self):
    >>>         return self._hello is not None

    Default values::
    >>> class A(object):
    >>>     int0 = field('int0', type=lambda: int32)
    >>>     has_int0 = int0.has_property
    >>> a = A()
    >>> a.int0 == 0
    >>> a.has_int0 is False
    '''
    def __init__(self, name, type0, is_discriminator=False):
        self.name = name
        self.private_name = '_' + name

        self._type_supplier = _supplier(type0)
        self._type = None
        self.is_discriminator = is_discriminator

    def __str__(self):
        return self.name + ' ' + self.type

    @property
    def type(self):
        '''Return field type descriptor.'''
        type0 = self._type
        if type0:
            return type0

        self._type = self._type_supplier()
        return self._type

    def __get__(self, message, owner=None):
        '''Get this field value in a message, check the type of the value.'''
        if message is None:
            # The field is accessed as a class attribute.
            # For example, TestMessage.field.
            return self

        value = getattr(message, self.private_name)
        if value is not None:
            return value

        default = self.type.default
        if self.type.is_mutable:
            setattr(message, self.private_name, default)

        return default

    def __set__(self, message, value):
        '''Set this field in a message to a value, check the type of the value.'''
        setattr(message, self.private_name, value)

    def __has__(self, message):
        '''Return True if this field is not None is a message.'''
        value = getattr(message, self.private_name)
        return value is not None

    @property
    def has_property(self):
        return property(self.__has__)


class InterfaceDescriptor(Descriptor):
    '''Interface descriptor.'''
    def __init__(self, pyclass, base=None, exc=None, methods=None):
        super(InterfaceDescriptor, self).__init__(Type.INTERFACE, pyclass)
        self.base = base
        self._exc_supplier = _supplier(exc)
        self._exc = None

        self.declared_methods = tuple(methods) if methods else ()
        self.inherited_methods = base.methods if base else ()
        self.methods = self.inherited_methods + self.declared_methods

    def __str__(self):
        return str(self.pyclass)

    @property
    def exc(self):
        if self._exc is None:
            self._exc = self._exc_supplier() if self._exc_supplier else None

        return self._exc or (self.base.exc if self.base else None)

    def find_method(self, name):
        '''Return a method by its name or None.'''
        for method in self.methods:
            if method.name == name:
                return method


class MethodDescriptor(object):
    '''Interface method descriptor.'''
    def __init__(self, name, result, args=None, is_post=False):
        self.name = name
        self._result_supplier = _supplier(result)
        self._result = None
        self.args = tuple(args) if args else ()

        self.is_post = is_post

    @property
    def result(self):
        '''Return a result descriptor.'''
        result = self._result
        if result:
            return result

        self._result = self._result_supplier()
        return self._result

    @property
    def is_terminal(self):
        '''Method is terminal when its result is not an interface.'''
        return self.result.type != Type.INTERFACE

    def invoke(self, obj, *args, **kwargs):
        '''Invoke this method on an object with a given arguments, return the result'''
        return getattr(obj, self.name)(*args, **kwargs)

    def __str__(self):
        '''Return a method signature as a string.'''
        s = [self.name, '(']

        next_separator = ''
        for arg in self.args:
            s.append(next_separator)
            s.append(arg.name)
            s.append(' ')
            s.append(str(arg.type))
            next_separator = ', '

        s.append(')=')
        s.append(str(self.result))

        return ''.join(s)


class ArgDescriptor(object):
    '''Method argument descriptor.'''
    def __init__(self, name, type0, is_query=False, is_post=False):
        self.name = name
        self._type_supplier = _supplier(type0)
        self._type = None

        self.is_query = is_query
        self.is_post = is_post

    @property
    def type(self):
        '''Return argument type descriptor.'''
        type0 = self._type
        if type0:
            return type0

        self._type = self._type_supplier()
        return self._type


class EnumDescriptor(DataTypeDescriptor):
    '''Enum descriptor.'''
    def __init__(self, pyclass, values):
        super(EnumDescriptor, self).__init__(Type.ENUM, pyclass)
        self.values = tuple(v.upper() for v in values)

    @property
    def default(self):
        return self.values[0]

    def find_value(self, name):
        if name is None:
            return None
        name = name.upper()

        if name not in self.values:
            return None
        return name


class ListDescriptor(DataTypeDescriptor):
    '''Internal list descriptor.'''
    def __init__(self, element):
        super(ListDescriptor, self).__init__(Type.LIST, list)
        self.element = element

    def __str__(self):
        return 'list<%s>' % self.element

    @property
    def default(self):
        return []


class SetDescriptor(DataTypeDescriptor):
    '''Internal set descriptor.'''
    def __init__(self, element):
        super(SetDescriptor, self).__init__(Type.SET, set)
        self.element = element

    def __str__(self):
        return 'set<%s>' % self.element

    @property
    def default(self):
        return set()


class MapDescriptor(DataTypeDescriptor):
    '''Internal map/dict descriptor.'''
    def __init__(self, key, value):
        super(MapDescriptor, self).__init__(Type.MAP, dict)
        self.key = key
        self.value = value

    def __str__(self):
        return 'map<%s, %s>' % (self.key, self.value)

    @property
    def default(self):
        return {}


class _PrimitiveDescriptor(DataTypeDescriptor):
    def __init__(self, type0, pyclass, default):
        super(_PrimitiveDescriptor, self).__init__(type0, pyclass)
        self.default = default


def list0(element):
    '''Create a list descriptor with an element descriptor.'''
    return ListDescriptor(element)


def set0(element):
    '''Create a set descriptor with an element descriptor.'''
    return SetDescriptor(element)


def map0(key, value):
    '''Create a map (dict) descriptor with key/value descriptors.'''
    return MapDescriptor(key, value)


def message(pyclass, base=None, discriminator_value=None, fields=None, subtypes=None):
    '''Create a message descriptor.'''
    return MessageDescriptor(pyclass, base=base, discriminator_value=discriminator_value,
                             fields=fields, subtypes=subtypes)


def field(name, type0, is_discriminator=False):
    '''Create a field descriptor.'''
    return FieldDescriptor(name, type0, is_discriminator=is_discriminator)


def enum(pyclass, values):
    '''Create an enum descriptor.'''
    return EnumDescriptor(pyclass, values)


def interface(pyclass, base=None, exc=None, methods=None):
    '''Create an interface descriptor.'''
    return InterfaceDescriptor(pyclass, base=base, exc=exc, methods=methods)


def method(name, result, args=None, is_post=False):
    '''Create an interface method descriptor.'''
    return MethodDescriptor(name, result=result, args=args, is_post=is_post)


def arg(name, type0, is_query=False, is_post=False):
    '''Create a method argument descriptor.'''
    return ArgDescriptor(name, type0, is_query=is_query, is_post=is_post)


def _supplier(type_or_lambda):
    if _is_lambda(type_or_lambda):
        # It is already a supplier.
        return type_or_lambda

    return lambda: type_or_lambda


def _is_lambda(type_or_lambda):
    if type_or_lambda is None:
        return None

    lambda0 = lambda: None
    lambda_type = type(lambda0)
    return isinstance(type_or_lambda, lambda_type) and type_or_lambda.__name__ == lambda0.__name__


bool0 = _PrimitiveDescriptor(Type.BOOL, bool, default=False)
int16 = _PrimitiveDescriptor(Type.INT16, int, default=0)
int32 = _PrimitiveDescriptor(Type.INT32, int, default=0)
int64 = _PrimitiveDescriptor(Type.INT64, _long, default=0)
float0 = _PrimitiveDescriptor(Type.FLOAT, float, default=0.0)
double0 = _PrimitiveDescriptor(Type.DOUBLE, float, default=0.0)
string0 = _PrimitiveDescriptor(Type.STRING, _string, default='')
datetime0 = _PrimitiveDescriptor(Type.DATETIME, _datetime, default=None)
void = _PrimitiveDescriptor(Type.VOID, type(None), default=None)



# encoding: utf-8
from __future__ import absolute_import
import types
import requests
import sys
import pdef
import pdef.descriptors
from pdef.invoke import Invocation
from pdefc.lang import TypeEnum

try:
    # Python 2.7
    import httplib as http_codes
except ImportError:
    # Python 3
    import http.client as http_codes


GET = 'GET'
POST = 'POST'
UTF8 = 'utf-8'
APPLICATION_JSON_CONTENT_TYPE = 'application/json; charset=utf-8'
FORM_URLENCODED_MIME_TYPE = 'application/x-www-form-urlencoded'
TEXT_PLAIN_CONTENT_TYPE = 'text/plain; charset=utf-8'


def rpc_client(interface, url, session=None):
    '''Create an RPC client.'''
    return RpcClient(interface, url, session=session)


def rpc_handler(interface, service):
    '''Create an RPC handler.'''
    return RpcHandler(interface, service)


def wsgi_app(handler):
    '''Create a WSGI RPC server.'''
    return WsgiRpcApp(handler)


class RpcException(Exception):
    def __init__(self, status, message=None):
        super(RpcException, self).__init__(message)
        self.message = message
        self.status = status


class RpcRequest(object):
    def __init__(self, method=GET, path='', query=None, post=None):
        self.method = method
        self.path = path
        self.query = dict(query) if query else {}
        self.post = dict(post) if post else {}

    def __str__(self):
        return '%s %s' % (self.method, self.path)

    def __repr__(self):
        return '<RpcRequest %s %s>' % (self.method, self.path)

    @property
    def is_post(self):
        return self.method == POST


class RpcProtocol(object):
    def __init__(self, jsonformat=None):
        self.jsonformat = jsonformat or pdef.jsonformat

    def get_request(self, invocation):
        if not invocation:
            raise ValueError('Invocation required')

        method = invocation.method
        if not method.is_terminal:
            raise ValueError('Last invocation method must be terminal')

        request = RpcRequest()
        request.method = POST if method.is_post else GET
        for inv in invocation.to_chain():
            self._write_invocation(request, inv)

        return request

    def _write_invocation(self, request, invocation):
        method = invocation.method
        kwargs = invocation.kwargs

        post = request.post
        query = request.query
        path = '/' + method.name

        for argd in method.args:
            name = argd.name
            kwarg = kwargs.get(name)
            if kwarg is None:
                if argd.is_post or argd.is_query:
                    continue
                raise ValueError('Path method argument "%s" cannot be null' % name)

            value = self._to_json(kwarg, argd.type)
            if argd.is_post:
                post[name] = value
            elif argd.is_query:
                query[name] = value
            else:
                path += '/' + urlencode(value)

        request.path += path

    def _to_json(self, kwarg, descriptor):
        '''Serialize a kwarg to json, strip quotes.'''
        s = self.jsonformat.write(kwarg, descriptor)
        return s.strip('"')

    def get_invocation(self, request, interface_descriptor):
        '''Parse an invocation from an rpc request using an interface descriptor.'''
        if not request:
            raise ValueError('Request required')
        if not interface_descriptor:
            raise ValueError('Interface descriptor required')

        invocation = None
        parts = request.path.strip('/').split('/')

        while parts:
            part = parts.pop(0)

            # Find a method by a name.
            method = interface_descriptor.find_method(part)
            if not method:
                raise RpcException(http_codes.BAD_REQUEST, 'Method not found')

            # Check the required HTTP method.
            if method.is_post and not request.is_post:
                raise RpcException(http_codes.METHOD_NOT_ALLOWED, 'Method not allowed, POST required')

            # Parse keyword arguments.
            kwargs = self._read_kwargs(method, parts, request.query, request.post)

            # Create a root invocation,
            # or a next invocation in a chain.
            invocation = invocation.next(method, **kwargs) \
                if invocation else Invocation(method, kwargs=kwargs)

            if method.is_terminal:
                break

            # It's an interface method.
            # Get the next interface descriptor and proceed parsing the parts.
            interface_descriptor = method.result

        if parts:
            # No more interface descriptors in a chain, but the parts are still present.
            raise RpcException(http_codes.BAD_REQUEST, 'Failed to parse an invocation chain')

        if not invocation:
            raise RpcException(http_codes.BAD_REQUEST, 'Methods required')

        if not invocation.method.is_terminal:
            raise RpcException(http_codes.BAD_REQUEST, 'The last method must be a terminal one. '
                                                       'It must return a data type or be void.')

        return invocation

    def _read_kwargs(self, method, parts, query, post):
        kwargs = {}

        for argd in method.args:
            name = argd.name

            if argd.is_post:
                value = post.get(name)
            elif argd.is_query:
                value = query.get(name)
            elif not parts:
                raise RpcException(http_codes.NOT_FOUND, 'Wrong number of method args: "%s"'
                                                      % method.name)
            else:
                value = urldecode(parts.pop(0))

            if value is None:
                kwargs[name] = None
            else:
                kwargs[name] = self._from_json(value, argd.type)

        return kwargs

    def _from_json(self, s, descriptor):
        if s is None:
            return None

        type0 = descriptor.type
        if type0 in (TypeEnum.STRING, TypeEnum.DATETIME, TypeEnum.ENUM):
            if not (s.startswith('"') and s.endswith('"')):
                # Return the quotes to get a valid json string.
                s = '"' + s + '"'

        return self.jsonformat.read(s, descriptor)


class RpcClient(object):
    def __init__(self, interface, url, session=None, protocol=None):
        if not interface:
            raise ValueError('Interface required')
        if not url:
            raise ValueError('Url required')

        self.interface = interface
        self.interface_descriptor = interface.descriptor

        self.url = url
        self.session = session or requests.session()
        self.protocol = protocol or RpcProtocol()

    def proxy(self):
        return pdef.proxy(self.interface, self)

    def __call__(self, invocation):
        if not invocation:
            raise ValueError('Invocation required')

        rpc_request = self.protocol.get_request(invocation)

        method = invocation.method
        resultd = method.result
        excd = self.interface_descriptor.exc

        request = self._build_request(rpc_request)
        return self._send(request, resultd, excd)

    def _build_request(self, rpc_request):
        url = self._build_url(rpc_request.path)
        return requests.Request(method=rpc_request.method,
                                url=url,
                                data=rpc_request.post,
                                params=rpc_request.query)

    def _build_url(self, path):
        return self.url + path

    def _send(self, request, resultd, excd=None):
        session = self.session
        prepared = request.prepare()
        response = session.send(prepared)
        return self._parse_response(response, resultd, excd)

    def _parse_response(self, response, resultd, excd=None):
        code = response.status_code

        if code not in (http_codes.OK, http_codes.UNPROCESSABLE_ENTITY):
            # It's an HTTP error.
            return self._parse_error(response)

        # It's a successful rpc result.
        text = response.text

        # Create a generic rpc result class.
        result_class = rpc_result_class(resultd, excd)
        result = result_class.from_json(text)

        if code == http_codes.OK:
            return result.data
        else:
            exc = result.error or RpcException(code, 'Unsupported application exception')
            raise exc

    def _parse_error(self, response):
        try:
            text = response.text
        except Exception as e:
            text = 'Failed to get the response text, e=%s' % e

        # Limit the text to use in as an exception description.
        text = text if len(text) < 255 else text[:255]
        raise RpcException(response.status_code, text)


class RpcHandler(object):
    def __init__(self, interface, service, protocol=None):
        if not interface:
            raise ValueError('Interface required')
        if not service:
            raise ValueError('Service required')
        self.interface = interface
        self.interface_descriptor = interface.descriptor
        self.service = service
        self.protocol = protocol or RpcProtocol()

    def __call__(self, rpc_request):
        return self.handle(rpc_request)

    def handle(self, rpc_request):
        '''Handle an rpc request and return a tuple (is_successful, rpc_result).'''
        if not rpc_request:
            raise ValueError('Rpc request required')

        invocation = self.protocol.get_invocation(rpc_request, self.interface_descriptor)

        method = invocation.method
        datad = method.result
        excd = self.interface_descriptor.exc
        result_class = rpc_result_class(datad, excd)

        try:
            data = invocation.invoke(self.service)
            return True, result_class(data)
        except Exception as e:
            if excd and isinstance(e, excd.pyclass):
                # It's an expected application exception.
                return False, result_class(data=None, error=e)

            # Not an application exception, reraise it.
            raise


class WsgiRpcApp(object):
    '''WSGI RPC application.'''
    def __init__(self, handler):
        if not handler:
            raise ValueError('Handler required')
        self.handler = handler

    def __call__(self, environ, start_response):
        return self.handle(environ, start_response)

    def handle(self, environ, start_response):
        request = self._parse_request(environ)
        try:
            success, result = self.handler(request)
        except RpcException as e:
            status = e.status or http_codes.INTERNAL_SERVER_ERROR
            content = e.message or 'Internal server error'
            return self._response(start_response, status, content)

        status_code = http_codes.OK if success else http_codes.UNPROCESSABLE_ENTITY
        content = result.to_json(indent=True)
        return self._response(start_response, status_code, content,
                              content_type=APPLICATION_JSON_CONTENT_TYPE)

    def _parse_request(self, env):
        '''Create an http server request from a wsgi request.'''
        method = env['REQUEST_METHOD']
        path = env['PATH_INFO']
        query = self._read_wsgi_query(env)
        post = self._read_wsgi_post(env)

        return RpcRequest(method, path=path, query=query, post=post)

    def _read_wsgi_query(self, env):
        if 'QUERY_STRING' not in env:
            return {}
        s = env['QUERY_STRING']
        return self._parse_query(s)

    def _read_wsgi_post(self, env):
        ctype = env.get('CONTENT_TYPE', '')
        clength = self._read_wsgi_clength(env)

        body = None
        if clength > 0 and ctype.lower().startswith(FORM_URLENCODED_MIME_TYPE):
            body = env['wsgi.input'].read(clength)

        if not body:
            return {}

        if sys.version > '3':
            body = body.decode(UTF8)

        return self._parse_query(body)

    def _read_wsgi_clength(self, env):
        clength = env.get('CONTENT_LENGTH') or 0
        try:
            return int(clength)
        except (ValueError, TypeError):
            return 0

    def _parse_query(self, s):
        d = parse_query(s)
        result = {}
        for key, values in d.items():
            result[key] = values[0] if values else ''
        return result

    def _response(self, start_response, status_code, unicode_content, content_type=None):
        reason = http_codes.responses.get(status_code)
        status = '%s %s' % (status_code,  reason)

        content = unicode_content.encode(UTF8)
        content_type = content_type or TEXT_PLAIN_CONTENT_TYPE
        headers = [('Content-Type', content_type),
                   ('Content-Length', str(len(content)))]

        start_response(status, headers)
        return [content]


def rpc_result_class(datad, excd=None):
    '''Create a generic RpcResult class with a given data and exception descriptors.'''
    class RpcResult(pdef.Message):
        data = pdef.descriptors.field('data', datad)
        error = pdef.descriptors.field('error', excd or pdef.descriptors.string0)
        descriptor = pdef.descriptors.message(lambda: RpcResult, fields=[data, error])

        has_data = data.has_property
        has_error = error.has_property

        def __init__(self, data=None, error=None):
            self.data = data
            self.error = error

    return RpcResult


try:
    # Python 2.7
    import urllib
    import urlparse

    def urlencode(s, safe='[]{},.-"'):
        if isinstance(s, types.UnicodeType):
            s = s.encode(UTF8)

        if isinstance(safe, types.UnicodeType):
            safe = safe.encode(UTF8)

        result = urllib.quote_plus(s, safe)
        return result.decode(UTF8)

    def urldecode(s):
        if isinstance(s, types.UnicodeType):
            s = s.encode(UTF8)

        result = urllib.unquote_plus(s)
        return result.decode(UTF8)

    def parse_query(s):
        if isinstance(s, types.UnicodeType):
            s = s.encode(UTF8)

        q = urlparse.parse_qs(s)

        # Decode the result to unicode.
        result = {}
        for key, values in q.items():
            k = key.decode(UTF8)
            vv = [v.decode(UTF8) for v in values] if values else []
            result[k] = vv

        return result

except ImportError:
    # Python 3
    import urllib.parse

    def urlencode(s, safe='[]{},.-"'):
        return urllib.parse.quote_plus(s, safe, encoding=UTF8)

    def urldecode(s):
        return urllib.parse.unquote_plus(s, encoding=UTF8)

    def parse_query(s):
        return urllib.parse.parse_qs(s)



# encoding: utf-8

# Copyright: 2013 Ivan Korobkov <ivan.korobkov@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pdef.types import Type, Message, Exc, Enum, Interface
from pdef.invoke import proxy
from pdef.formats import jsonformat
from pdef.rpc import rpc_client, rpc_handler, wsgi_app
from pdef.version import __version__

__title__ = 'pdef'
__author__ = 'Ivan Korobkov <ivan.korobkov@gmail.com>'
__license__ = 'Apache License 2.0'
__copyright__ = 'Copyright 2013 Ivan Korobkov'



# encoding: utf-8
from __future__ import unicode_literals
import copy
import sys
import pdef


class Type(object):
    '''Pdef type enum.'''

    # Primitive types
    BOOL = 'bool'
    INT16 = 'int16'
    INT32 = 'int32'
    INT64 = 'int64'
    FLOAT = 'float'
    DOUBLE = 'double'
    STRING = 'string'
    DATETIME = 'datetime'

    # Void
    VOID = 'void'

    # Collection types.
    LIST = 'list'
    MAP = 'map'
    SET = 'set'

    # User defined types.
    ENUM = 'enum'
    MESSAGE = 'message'
    INTERFACE = 'interface'

    PRIMITIVE_TYPES = (BOOL, INT16, INT32, INT64, FLOAT, DOUBLE, STRING)
    DATA_TYPES = PRIMITIVE_TYPES + (DATETIME, LIST, MAP, SET, ENUM, MESSAGE, VOID)
    ALL_TYPES = DATA_TYPES + (INTERFACE, )
    MUTABLE_TYPES = (LIST, MAP, SET, MESSAGE)


class Enum(object):
    descriptor = None


class Interface(object):
    descriptor = None


class Message(object):
    descriptor = None

    @classmethod
    def from_json(cls, s, **kwargs):
        '''Parse a message from a json string.'''
        return pdef.jsonformat.read(s, cls.descriptor, **kwargs)

    @classmethod
    def from_json_stream(cls, fp, **kwargs):
        '''Parse a message from a json file-like object.'''
        return pdef.jsonformat.read_stream(fp, cls.descriptor, **kwargs)

    @classmethod
    def from_dict(cls, d):
        '''Parse a message from a dictionary.'''
        return pdef.jsonformat.read_object(d, cls.descriptor)

    def to_json(self, indent=None, **kwargs):
        '''Convert this message to a json string.'''
        return pdef.jsonformat.write(self, self.descriptor, indent=indent)

    def to_json_stream(self, fp, indent=None, **kwargs):
        '''Serialize this message as a json string to a file-like stream.'''
        return pdef.jsonformat.write_to_stream(self, self.descriptor, fp, indent=indent, **kwargs)

    def to_dict(self):
        '''Convert this message to a dictionary (serialize each field).'''
        return pdef.jsonformat.write_object(self, self.descriptor)

    def merge(self, message):
        '''Deep copy present fields from another message into this one.'''
        if not message:
            return

        descriptor = self.descriptor
        if not isinstance(message, self.__class__):
            if not isinstance(self, message.__class__):
                return
            descriptor = message.descriptor

        for field in descriptor.fields:
            if field.is_discriminator:
                continue

            value = getattr(message, field.private_name)
            if value is None:
                continue

            value_copy = copy.deepcopy(value)
            setattr(self, field.name, value_copy)

        return self

    def merge_dict(self, d):
        '''Parse a message from a dict and merge it into this message.'''
        message = self.__class__.from_dict(d)
        return self.merge(message)

    def merge_json(self, s):
        '''Parse a message from a json string and merge it into this message.'''
        message = self.__class__.from_json(s)
        return self.merge(message)

    def merge_json_stream(self, stream):
        '''Parse a message from a file-like object with json data and merge it into this message.'''
        message = self.__class__.from_json_stream(stream)
        return self.merge(message)

    def __eq__(self, other):
        if other is None or self.__class__ is not other.__class__:
            return False
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not self == other

    def __copy__(self):
        msg = self.__class__()
        msg.__dict__ = copy.copy(self.__dict__)
        return msg

    def __deepcopy__(self, memo=None):
        msg = self.__class__()
        msg.__dict__ = copy.deepcopy(self.__dict__, memo)
        return msg

    def __str__(self):
        s = self.__unicode__()
        if sys.version < '3':
            return s.encode('utf-8', errors='replace')
        return s

    def __unicode__(self):
        s = ['<', self.__class__.__name__, ' ']
        to_unicode = str
        if sys.version < '3':
            to_unicode = unicode

        first = True
        d = self.__dict__
        for key, value in d.items():
            if first:
                first = False
            else:
                s.append(', ')

            s.append(key)
            s.append('=')
            s.append(to_unicode(value))

        s.append('>')
        return ''.join(s)


class Exc(Exception, Message):
    pass



# encoding: utf-8
from datetime import datetime
import json as _json
import pdef.types

SIMPLE_ISO_8601_PATTERN = "%Y-%m-%dT%H:%M:%SZ"


class JsonFormat(object):
    '''JsonFormat parses/serializes Pdef types from/to JSON.'''

    def __init__(self):
        self.object_format = _JsonObjectFormat()

    def read(self, s, descriptor):
        '''Read a pdef object from a JSON string.'''
        if s is None:
            return None

        value = _json.loads(s)
        parsed = self.object_format.read(value, descriptor)
        return parsed

    def read_stream(self, fp, descriptor):
        '''Read a pdef object from a a JSON file-like object.'''
        value = _json.load(fp)
        parsed = self.object_format.read(value, descriptor)
        return parsed

    def read_object(self, data, descriptor):
        '''Read a pdef object from a JSON-compatible object.'''
        return self.object_format.read(data, descriptor)

    def write(self, obj, descriptor, indent=None, **kwargs):
        '''Write a pdef object to a JSON string.'''
        serialized = self.object_format.write(obj, descriptor)
        s = _json.dumps(serialized, ensure_ascii=False, indent=indent, **kwargs)
        return s

    def write_to_stream(self, obj, descriptor, fp, indent=None, **kwargs):
        '''Write a pdef object as a JSON string to a file-like object.'''
        serialized = self.object_format.write(obj, descriptor)
        return _json.dump(serialized, fp, ensure_ascii=False, indent=indent, **kwargs)

    def write_object(self, obj, descriptor):
        '''Write a pdef object into a JSON-compatible object.'''
        return self.object_format.write(obj, descriptor)


class _JsonObjectFormat(object):
    '''JsonObjectFormat parses/serializes Pdef objects from/to JSON-compatible objects.'''

    def write(self, obj, descriptor):
        '''Write an object to a native python type.'''
        if obj is None:
            return None

        type0 = descriptor.type
        if type0 in pdef.types.Type.PRIMITIVE_TYPES:
            # This is for type checks.
            return descriptor.pyclass(obj)

        write = self.write
        write_key = self._write_key
        Type = pdef.types.Type

        if type0 == Type.DATETIME:
            if not isinstance(obj, datetime):
                raise ValueError('Not a datetime object %r' % datetime)
            return obj.strftime(SIMPLE_ISO_8601_PATTERN)

        elif type0 == Type.ENUM:
            return obj.lower()

        elif type0 == Type.LIST:
            elemd = descriptor.element
            return [write(elem, elemd) for elem in obj]

        elif type0 == Type.SET:
            elemd = descriptor.element
            return [write(elem, elemd) for elem in obj]

        elif type0 == Type.MAP:
            keyd = descriptor.key
            valued = descriptor.value
            return {write_key(k, keyd): write(v, valued) for k, v in obj.items() if k is not None}

        elif type0 == Type.MESSAGE:
            return self._message_to_dict(obj)

        elif type0 == Type.VOID:
            return None

        raise ValueError('Unsupported type ' + descriptor)

    def _write_key(self, key, keyd):
        if key is None:
            raise ValueError('None dict key')

        type0 = keyd.type
        Type = pdef.types.Type

        if type0 == Type.BOOL:
            return 'true' if key else 'false'

        elif type0 in Type.PRIMITIVE_TYPES:
            return str(key)

        elif type0 == Type.DATETIME:
            return key.strftime(SIMPLE_ISO_8601_PATTERN)

        raise ValueError('Unsupported key type ' + keyd)

    def _message_to_dict(self, message):
        if message is None:
            return None

        result = {}
        to_object = self.write
        descriptor = message.descriptor  # Support polymorphic messages.

        # field_tuples and private names are a performance optimization.
        # It's about 30% faster.
        for name, private_name, type0 in descriptor.field_tuples:
            value = getattr(message, private_name)
            if value is None:
                continue

            result[name] = to_object(value, type0)
        return result

    def read(self, data, descriptor):
        '''Read a pdef object from a native python type.'''
        if data is None:
            return None

        type0 = descriptor.type
        Type = pdef.types.Type
        read = self.read

        if type0 in Type.PRIMITIVE_TYPES:
            return descriptor.pyclass(data)

        elif type0 == Type.DATETIME:
            if isinstance(data, datetime):
                return data
            return datetime.strptime(data, SIMPLE_ISO_8601_PATTERN)

        elif type0 == Type.ENUM:
            return descriptor.find_value(data)

        elif type0 == Type.LIST:
            elemd = descriptor.element
            return [read(elem, elemd) for elem in data]

        elif type0 == Type.SET:
            elemd = descriptor.element
            return {read(elem, elemd) for elem in data}

        elif type0 == Type.MAP:
            keyd = descriptor.key
            valued = descriptor.value
            return {read(k, keyd): read(v, valued) for k, v in data.items()}

        elif type0 == Type.MESSAGE:
            return self._message_from_dict(data, descriptor)

        elif type0 == Type.VOID:
            return None

        raise ValueError('Unsupported type ' + descriptor)

    def _message_from_dict(self, dict0, descriptor):
        '''Parse a message from a dictionary.'''
        if dict0 is None:
            return None

        from_object = self.read

        if descriptor.is_polymorphic:
            # Parse a discriminator value and find a subtype descriptor.
            discriminator = descriptor.discriminator
            serialized = dict0.get(discriminator.name)
            parsed = from_object(serialized, discriminator.type)
            descriptor = descriptor.find_subtype(parsed)

        message = descriptor.pyclass()
        for field in descriptor.fields:
            serialized = dict0.get(field.name)
            if serialized is None:
                continue

            parsed = from_object(serialized, field.type)
            setattr(message, field.name, parsed)

        return message


jsonformat = JsonFormat()



# encoding: utf-8
import copy
from collections import deque


def proxy(interface, invocation_handler):
    '''Create a interface proxy.

    The proxy captures method calls into chained invocations, and passes them
    to the handler on terminal methods.

    @param interface:               Interface with a class descriptor field.
    @param invocation_handler:      callable(Invocation): InvocationResult.
    '''
    descriptor = interface.descriptor
    return InvocationProxy(descriptor, invocation_handler)


class Invocation(object):
    '''Immutable chained method invocation.'''
    @classmethod
    def root(cls, method, *args, **kwargs):
        return Invocation(method, args=args, kwargs=kwargs)

    def __init__(self, method, args=None, kwargs=None, parent=None):
        '''Create an invocation.

        @param method: a method descriptor.
        @param parent: a parent invocation, nullable.
        '''
        self.method = method
        self.parent = parent

        self.kwargs = self._build_kwargs(method, args, kwargs) if method else {}
        self.kwargs = copy.deepcopy(self.kwargs)    # Make defensive copies.

    def __repr__(self):
        return '<Invocation %r args=%r>' % (self.method.name, self.kwargs)

    @property
    def kwargs_with_default_primitives(self):
        result = {}
        kwargs = self.kwargs

        for argd in self.method.args:
            arg = kwargs.get(argd.name)
            if arg is None and argd.type.is_primitive:
                arg = argd.type.default
            result[argd.name] = arg

        return result

    def next(self, method, *args, **kwargs):
        '''Create a child invocation.'''
        return Invocation(method, args=args, kwargs=kwargs, parent=self)

    def to_chain(self):
        '''Return a list of invocations.'''
        if not self.parent:
            return [self]

        chain = self.parent.to_chain()
        chain.append(self)
        return chain

    def invoke(self, obj):
        '''Invoke this invocation chain on an object.'''
        chain = self.to_chain()

        for inv in chain:
            obj = inv.method.invoke(obj, **inv.kwargs_with_default_primitives)

        return obj

    @staticmethod
    def _build_kwargs(method, args=None, kwargs=None):
        '''Convert args and kwargs into a param dictionary, check their number and types.'''
        def wrong_args():
            return TypeError('Wrong method arguments, %s, got args=%s, kwargs=%s' %
                             (method, args, kwargs))
        params = {}
        args = args if args else ()
        kwargs = kwargs if kwargs else {}
        method_args = deque(method.args)

        # Check that the number of args and kwargs is less or equal to the method args number.
        if len(method_args) < (len(args) + len(kwargs)):
            raise wrong_args()

        # Consume all positional arguments.
        for value in args:
            argd = method_args.popleft()
            params[argd.name] = value

        # Add keyword arguments using the remaining method args.
        consumed_kwargs = {}
        for argd in method_args:
            if argd.name not in kwargs:
                params[argd.name] = None
                continue

            value = kwargs.get(argd.name)
            params[argd.name] = value
            consumed_kwargs[argd.name] = value

        # Check that all kwargs have been consumed.
        if list(consumed_kwargs.keys()) != list(kwargs.keys()):
            raise wrong_args()

        return params


class InvocationProxy(object):
    '''Reflective client proxy.'''
    def __init__(self, descriptor, handler, invocation=None):
        self._descriptor = descriptor
        self._handler = handler
        self._invocation = invocation

    def __repr__(self):
        return '<InvocationProxy %s>' % self._descriptor

    def __getattr__(self, name):
        '''Return a proxy method.'''
        method = self._descriptor.find_method(name)
        if not method:
            raise AttributeError('Method not found %r' % name)

        return _ProxyMethod(self._invocation, self._handler, method)


class _ProxyMethod(object):
    def __init__(self, invocation, handler, method):
        self.invocation = invocation
        self.handler = handler
        self.method = method

    def __repr__(self):
        return '<ProxyMethod %r>' % self.method.name

    def __call__(self, *args, **kwargs):
        method = self.method
        if self.invocation:
            invocation = self.invocation.next(method, *args, **kwargs)
        else:
            invocation = Invocation(method, args=args, kwargs=kwargs)

        if not method.is_terminal:
            # This is a method, which returns an interface.
            # Create a next invocation proxy.
            return InvocationProxy(method.result, self.handler, invocation)

        # The method result is a value or void.
        result = self.handler(invocation)
        if result is not None:
            return result

        if method.result.is_primitive:
            return method.result.default
        return None



# encoding: utf-8
import os.path
try:
    from setuptools import setup
except ImportError:
    import ez_setup
    ez_setup.use_setuptools()
    from setuptools import setup


# Load the version.
__version__ = None
with open(os.path.join('src', 'pdef', 'version.py')) as f:
    exec(f.read())


setup(
    name='pdef',
    version=__version__,
    license='Apache License 2.0',
    description='Protocol definition language',
    long_description=open('README.md', 'r').read(),
    url='http://github.com/pdef/pdef-python',

    author='Ivan Korobkov',
    author_email='ivan.korobkov@gmail.com',

    packages=['pdef'],
    package_dir={'': 'src'},
    py_modules=['pdef.rpc'],

    install_requires=['requests>=1.2'],

    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: WSGI',
        'Topic :: Software Development :: Libraries'
    ]
)
